﻿using System;                                 // needed for event stuff
using System.Collections;                     // needed for ArrayList collection
using System.Drawing;                         // needed for Color structure
using Microsoft.DirectX;                      // needed for vector stuff
using Direct3D = Microsoft.DirectX.Direct3D;  // needed for Direct3D
using VDEngine.Game;
using VDEngine.Utility;

namespace VDEngine.DirectX
{
    /// <summary>
    /// Represents everything that is done in Direct3D.
    /// </summary>
    public class VDGraphics
    {
        private VDEngine _engine;                                    // the game engine
        private int _adapter;                                        // device adapter
        private Direct3D.Format _displayFormat;                      // display format
        private Direct3D.CreateFlags _behaviorFlags;                 // behavior flags of device
        private Direct3D.PresentParameters _presentationParameters;  // presentation parameters of device
        private Direct3D.Device _device;                             // device itself
        private bool _deviceLost;                                    // device lost
        private bool _deviceReset;                                   // device reset
        private Direct3D.Material _material;                         // material (diffuse, ambient, emissive, specular)
        private Vector3 _worldTransformation;                        // world transformation

        private VDTexture _testTexture;               // TODO: temporary (for testing a texture)
        private Direct3D.Font _tempText;              // TODO: temporary (for showing FPS)
        private Direct3D.Texture _testSpriteTexture;  // TODO: temporary (for testing a textured point sprite)

        /// <summary>
        /// Mouse movement.
        /// </summary>
        public enum MouseLook
        {
            /// <summary>
            /// Mouse isn't moving.
            /// </summary>
            Stasis,
            /// <summary>
            /// Positive Y axis.
            /// </summary>
            Up,
            /// <summary>
            /// Negative Y axis.
            /// </summary>
            Down,
            /// <summary>
            /// Negative X axis.
            /// </summary>
            Left,
            /// <summary>
            /// Positive X axis.
            /// </summary>
            Right
        }



        #region Initialization (Constructor)

        /// <summary>
        /// Initializes the display.
        /// </summary>
        public VDGraphics(VDEngine engine)
        {
            _engine = engine;                                      // make sure we know about the game engine
            _worldTransformation = new Vector3(0.0f, 0.0f, 0.0f);  // world transformation

            // figure out everything the user's computer is capable of
            DetermineOptions();

            // initialize the presentation parameters
            InitializePresentationParameters();

            // initialize Direct3D
            InitializationDevice();

            // initialize the camera
            InitializeCamera();

            // temporary things (TODO)
            _testSpriteTexture = Direct3D.TextureLoader.FromFile(_device, "../../../sprite.bmp");         // point sprite
            _testTexture = new VDTexture(_device, _engine.DisplaySettings.AspectRatio);                // texture
            System.Drawing.Font newFont = new System.Drawing.Font("Arial", 12.0f, FontStyle.Regular);  // FPS display
            _tempText = new Direct3D.Font(_device, newFont);
        }

        #endregion



        #region Direct3D Initialization

        /// <summary>
        /// Finds the user's available display options
        /// </summary>
        private void DetermineOptions()
        {
            Direct3D.Caps hardwareCapabilities;  // hardware capabilities

            // get the default adapter
            _adapter = Direct3D.Manager.Adapters.Default.Adapter;

            // get the hardware capabilities
            hardwareCapabilities = Direct3D.Manager.GetDeviceCaps(_adapter, Direct3D.DeviceType.Hardware);

            // check for hardware or software vertex processing support
            if (hardwareCapabilities.DeviceCaps.SupportsHardwareTransformAndLight)
            {
                _behaviorFlags = Direct3D.CreateFlags.HardwareVertexProcessing;
            }
            else
            {
                _behaviorFlags = Direct3D.CreateFlags.SoftwareVertexProcessing;
            }

            // check for pure device support
            if (hardwareCapabilities.DeviceCaps.SupportsPureDevice)
            {
                _behaviorFlags |= Direct3D.CreateFlags.PureDevice;
            }

            // set the display format to current one
            _displayFormat = Direct3D.Manager.Adapters.Default.CurrentDisplayMode.Format;
        }

        /// <summary>
        /// Initializes the presentation of Direct3D
        /// </summary>
        private void InitializePresentationParameters()
        {
            _presentationParameters = new Direct3D.PresentParameters();

            _presentationParameters.AutoDepthStencilFormat = Direct3D.DepthFormat.D16;             // 16-bit depth buffer
            _presentationParameters.BackBufferCount = 1;                                           // the number of back buffers
            _presentationParameters.BackBufferFormat = _displayFormat;                             // display format of the back buffer
            _presentationParameters.BackBufferHeight = _engine.DisplaySettings.WindowSize.Height;  // height of the back buffer
            _presentationParameters.BackBufferWidth = _engine.DisplaySettings.WindowSize.Width;    // width of the back buffer
            _presentationParameters.DeviceWindow = _engine;                                        // set the display window
            //_presentationParameters.DeviceWindowHandle;                                          // ignored (used in Win32)
            _presentationParameters.EnableAutoDepthStencil = true;                                 // set Direct3D to manage depth buffers
            _presentationParameters.ForceNoMultiThreadedFlag = false;                              // true if multithreading is allowed, false otherwise
            //_presentationParameters.FullScreenRefreshRateInHz;                                   // allow Direct3D to manage
            //_presentationParameters.MultiSample;                                                 // TODO: anti-aliasing
            //_presentationParameters.MultiSampleQuality;                                          // TODO: anti-aliasing
            _presentationParameters.PresentationInterval = Direct3D.PresentInterval.Immediate;     // Wait for VSync (off)
            //_presentationParameters.PresentFlag;                                                 // TODO: figure out if this shit matters
            _presentationParameters.SwapEffect = Direct3D.SwapEffect.Discard;                      // avoid copying/flipping overheads of the swap chain
            _presentationParameters.Windowed = !_engine.DisplaySettings.FullScreen;                // indicates whether the application is windowed or full-screen
        }

        /// <summary>
        /// Initializes Direct3D
        /// </summary>
        private void InitializationDevice()
        {
            // create the Direct3D device
            _device = new Direct3D.Device(_adapter,                      // adapter
                                          Direct3D.DeviceType.Hardware,  // device type
                                          _engine,                       // rendering window
                                          _behaviorFlags,                // behavior flags
                                          _presentationParameters);      // presentation parameters
            _deviceLost = false;
            _deviceReset = false;

            // turn off MDX automatic resize handling
            _device.DeviceResizing += new System.ComponentModel.CancelEventHandler(OnDeviceResize);

            // register the event handlers for lost devices
            _device.DeviceLost += new EventHandler(OnDeviceLost);
            _device.DeviceReset += new EventHandler(OnDeviceReset);

            // render states
            _device.RenderState.CullMode = Direct3D.Cull.None;  // back-facing triangles are not culled
            _device.RenderState.Lighting = false;               // disable Direct3D lighting

            // material color source (lighting calculations)
            _material = new Direct3D.Material();
            //_material.Ambient = ?;
            _material.Diffuse = Color.White;       // diffuse color
            //_material.Emissive = ?;
            _material.Specular = Color.LightGray;  // specular color
            _material.SpecularSharpness = 15.0F;   // sharpness of specular highlights
            _device.Material = _material;
        }

        /// <summary>
        /// Event for when the display device gets resized.
        /// </summary>
        /// <param name="sender">Sender of the event.  This parameter is not used.</param>
        /// <param name="cancelEventArguments">Cancel arguments of the event.</param>
        private void OnDeviceResize(object sender, System.ComponentModel.CancelEventArgs cancelEventArguments)
        {
            cancelEventArguments.Cancel = true;
        }

        /// <summary>
        /// Event for when the display device gets lost.
        /// </summary>
        /// <param name="sender">Sender of the event.</param>
        /// <param name="eventArguments">Event arguments.</param>
        private void OnDeviceLost(object sender, EventArgs eventArguments)
        {
            _deviceLost = true;
        }

        /// <summary>
        /// Event for when the display device gets reset.
        /// </summary>
        /// <param name="sender">Sender of the event.</param>
        /// <param name="eventArguments">Event arguments.</param>
        private void OnDeviceReset(object sender, EventArgs eventArguments)
        {
            _deviceReset = true;
        }

        /// <summary>
        /// Initializes the camera.
        /// </summary>
        private void InitializeCamera()
        {
            // transform 3-D view space coordinates into 2-D screen coordinates
            _device.Transform.Projection = Matrix.PerspectiveFovLH((float) Math.PI / 4.0f,                     // field of view in the y direction (radians)
                                                                   _engine.DisplaySettings.AspectRatio,        // aspect ratio (width / height)
                                                                   _engine.DisplaySettings.NearClippingPlane,  // near clipping plane (z-value)
                                                                   _engine.DisplaySettings.FarClippingPlane);  // far clipping plane (z-value)

            // point of view
            _device.Transform.View = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, 10.0f),  // camera's position
                                                     new Vector3(0.0f, 0.0f, 0.0f),   // camera's target is at the origin
                                                     new Vector3(0.0f, 1.0f, 0.0f));  // camera's up position is in the positive y direction
        }

        #endregion



        #region Rendering

        /// <summary>
        /// Calculates the new world transformation matrix.
        /// </summary>
        public void CalculateNewWorldView()
        {
            _worldTransformation.X = (float) Math.Sin(_engine.Player.Direction.X) + _engine.Player.Location.X;
            _worldTransformation.Y = (float) Math.Sin(_engine.Player.Direction.Y);
            _worldTransformation.Z = (float) Math.Cos(_engine.Player.Direction.X) + (float) Math.Cos(_engine.Player.Direction.Y) + _engine.Player.Location.Z;
        }

        /// <summary>
        /// This checks to make sure we can render, and then renders.
        /// </summary>
        public void Render()
        {
            int result;

            // check if the device is lost or needs to be reset
            if (_device.CheckCooperativeLevel(out result))
            {
                // try to render the screen
                try
                {
                    RenderGraphics();
                }
                catch (Direct3D.DeviceLostException)
                {
                    _device.CheckCooperativeLevel(out result);
                }
                catch (Direct3D.DeviceNotResetException)
                {
                    _device.CheckCooperativeLevel(out result);
                }
            }
            if (result == (int) Direct3D.ResultCode.DeviceLost)
            {
                // wait until we find the device
                return;
            }
            else if (result == (int) Direct3D.ResultCode.DeviceNotReset)
            {
                // reset the device
                _device.Reset(_presentationParameters);
            }
        }

        /// <summary>
        /// Renders the graphics.
        /// </summary>
        private void RenderGraphics()
        {
            // clear the viewport
            _device.Clear(Direct3D.ClearFlags.Target | Direct3D.ClearFlags.ZBuffer, Color.Black, 1.0f, 0);

            // begin the scene
            _device.BeginScene();

            // show the FPS
            _tempText.DrawText(null, string.Format("FPS: {0}", _engine.Timer.FramesPerSecond), new Point(400, 20), Color.White);

            // change the user's perspective
            _device.Transform.World = Matrix.Translation(_worldTransformation)
                                    * Matrix.RotationYawPitchRoll(-_engine.Player.Direction.X, _engine.Player.Direction.Y, 0.0f);

            // draw the test texture
            _device.VertexFormat = Direct3D.CustomVertex.PositionTextured.Format;
            _device.SetTexture(0, _testTexture.Texture);
            _device.DrawUserPrimitives(Direct3D.PrimitiveType.TriangleList, 2, _testTexture.Vertices);

            // draw the particle system
            _engine.ParticleSystem.UpdatePosition(_engine.Timer.ElapsedMilliseconds);
            _device.SetTexture(0, _testSpriteTexture);
            _device.Material = _material;
            _device.VertexFormat = Direct3D.VertexFormats.Position | Direct3D.VertexFormats.PointSize | Direct3D.VertexFormats.Diffuse;
            _engine.ParticleSystem.Render(_device);

            // end the scene
            _device.EndScene();

            // show the scene
            _device.Present();
        }

        #endregion



        #region Shutdown

        /// <summary>
        /// Releases the resources of the Direct3D graphics.
        /// </summary>
        public void Shutdown()
        {
            if (_testTexture.Texture != null)
            {
                _testTexture.Shutdown();
            }
            if (_device != null)
            {
                _device.Dispose();
                _device = null;
            }
        }

        #endregion



        #region Accessors & Modifiers

        /// <summary>
        /// Accessor for the graphics device.
        /// </summary>
        public Direct3D.Device Device
        {
            get
            {
                return _device;
            }
        }

        /// <summary>
        /// Accessor for the lost status of the graphics device.
        /// </summary>
        public bool DeviceLost
        {
            get
            {
                return _deviceLost;
            }
        }

        /// <summary>
        /// Accessor for the reset status of the graphics device.
        /// </summary>
        public bool DeviceReset
        {
            get
            {
                return _deviceReset;
            }
        }

        #endregion
    }
}
