﻿using Microsoft.DirectX.DirectInput;  // needed for DirectInput

namespace VDEngine.DirectX
{
    /// <summary>
    /// Gathers data from devices such as the keyboard and mouse.
    /// </summary>
    public class VDInput
    {
        private Device _keyboardDevice;            // DirectInput keyboard device
        private bool[] _keysActivated;             // key strokes
        private bool[] _keysHeldDown;              // keys being held down
        private Device _mouseDevice;               // DirectInput mouse device
        private int _previousMouseStateX;          // mouse's previous X coordinate
        private int _previousMouseStateY;          // mouse's previous Y coordinate
        private VDGraphics.MouseLook _mouseMoved;  // movement of the mouse
        private int _mouseSpeed;                   // speed of the mouse



        #region Initialization (Constructor)

        /// <summary>
        /// Initializes the input.
        /// </summary>
        /// <param name="parent"></param>
        public VDInput(System.Windows.Forms.Control parent)
        {
            _keysActivated = new bool[256];
            _keysHeldDown = new bool[256];
            _previousMouseStateX = 0;
            _previousMouseStateY = 0;
            _mouseMoved = VDGraphics.MouseLook.Stasis;
            _mouseSpeed = 0;

            // initialize the keyboard & mouse device
            _keyboardDevice = new Device(SystemGuid.Keyboard);
            _mouseDevice = new Device(SystemGuid.Mouse);

            // set the data format
            _keyboardDevice.SetDataFormat(DeviceDataFormat.Keyboard);
            _mouseDevice.SetDataFormat(DeviceDataFormat.Mouse);

            // set the cooperative level
            // Background: can be used in background and acquired at any time, even if the window is not active
            // NonExclusive: can be shared with any other applications not requiring Exclusive access
            _keyboardDevice.SetCooperativeLevel(parent, CooperativeLevelFlags.Background | CooperativeLevelFlags.NonExclusive);
            _mouseDevice.SetCooperativeLevel(parent, CooperativeLevelFlags.Background | CooperativeLevelFlags.NonExclusive);

            // try to access the keyboard & mouse
            try
            {
                _keyboardDevice.Acquire();
                _mouseDevice.Acquire();
            }
            catch (InputException)
            {
                // TODO: handle this better...
                //       maybe we should just fucking loop until it DOES acquire the devices?
            }
        }

        #endregion



        #region Updating Input Devices

        /// <summary>
        /// Updates the keyboard device.
        /// </summary>
        public void UpdateKeyboard()
        {
            try
            {
                Key[] currentPressedKeys = _keyboardDevice.GetPressedKeys();

                // update the key strokes

                // reset the key strokes to none
                for (int keyIndex = 0; keyIndex < _keysActivated.Length; keyIndex++)
                {
                    _keysActivated[keyIndex] = false;
                }

                // go through all the keys currently being pressed
                for (int keyIndex = 0; keyIndex < currentPressedKeys.Length; keyIndex++)
                {
                    int key = (int) currentPressedKeys[keyIndex];

                    // as long as the key's not being held down, it's a stroke
                    if (!_keysHeldDown[keyIndex])
                    {
                        _keysActivated[keyIndex] = true;
                    }
                }

                // update the held keys

                // reset the held keys
                for (int keyIndex = 0; keyIndex < _keysHeldDown.Length; keyIndex++)
                {
                    _keysHeldDown[keyIndex] = false;
                }

                // go through all the keys currently being pressed
                for (int keyIndex = 0; keyIndex < currentPressedKeys.Length; keyIndex++)
                {
                    int key = (int) currentPressedKeys[keyIndex];

                    // after we note the key as a stroke, note it as being held... so when the next call happens, it will not longer be a stroke
                    _keysHeldDown[key] = true;
                }
            }
            catch (InputException)
            {
                _keyboardDevice.Acquire();
            }
        }

        /// <summary>
        /// Updates the mouse device.
        /// </summary>
        public void UpdateMouse()
        {
            _previousMouseStateX = 0;
            _previousMouseStateY = 0;

            try
            {
                // poll the mouse just in case, and get the current state
                _mouseDevice.Poll();
                MouseState currentMouseState = _mouseDevice.CurrentMouseState;

                // check the various mouse possibilities
                if (currentMouseState.Y > _previousMouseStateY)
                {
                    _mouseMoved = VDGraphics.MouseLook.Up;
                    _mouseSpeed = currentMouseState.Y - _previousMouseStateY;
                }
                else if (currentMouseState.Y < _previousMouseStateY)
                {
                    _mouseMoved = VDGraphics.MouseLook.Down;
                    _mouseSpeed = currentMouseState.Y - _previousMouseStateY;
                }
                else if (currentMouseState.X < _previousMouseStateX)
                {
                    _mouseMoved = VDGraphics.MouseLook.Left;
                    _mouseSpeed = currentMouseState.X - _previousMouseStateX;
                }
                else if (currentMouseState.X > _previousMouseStateX)
                {
                    _mouseMoved = VDGraphics.MouseLook.Right;
                    _mouseSpeed = currentMouseState.X - _previousMouseStateX;
                }
                else
                {
                    _mouseMoved = VDGraphics.MouseLook.Stasis;
                    _mouseSpeed = 0;
                }
            }
            catch (InputException)
            {
                _mouseDevice.Acquire();
            }
        }

        #endregion



        #region Shutdown

        /// <summary>
        /// Releases the resources of the input devices.
        /// </summary>
        public void Shutdown()
        {
            if (_keyboardDevice != null)
            {
                _keyboardDevice.Unacquire();
                _keyboardDevice.Dispose();
                _keyboardDevice = null;
            }
            if (_mouseDevice != null)
            {
                _mouseDevice.Unacquire();
                _mouseDevice.Dispose();
                _mouseDevice = null;
            }
        }

        #endregion



        #region Accessors & Modifiers

        /// <summary>
        /// See if a key stroke has been made.
        /// </summary>
        /// <param name="key">The key pressed.</param>
        /// <returns>True if it was pressed, false otherwise.</returns>
        public bool KeyActivated(Key key)
        {
            return _keysActivated[(int) key];
        }

        /// <summary>
        /// See if a key is being held down.
        /// </summary>
        /// <param name="key">The key pressed.</param>
        /// <returns>True if it is being pressed, false otherwise.</returns>
        public bool KeyHeldDown(Key key)
        {
            return _keysHeldDown[(int) key];
        }

        /// <summary>
        /// Accessor to see if the mouse moved.
        /// </summary>
        public VDGraphics.MouseLook MouseMoved
        {
            get
            {
                return _mouseMoved;
            }
        }

        /// <summary>
        /// Accessor for the mouse speed.
        /// </summary>
        public int MouseSpeed
        {
            get
            {
                return _mouseSpeed;
            }
        }

        #endregion
    }
}
