﻿using System;
using System.Drawing;
using Microsoft.DirectX;
using Direct3D = Microsoft.DirectX.Direct3D;
using VDEngine.Game;

namespace VDEngine.DirectX
{
    /// <summary>
    /// Representation of a 2D object in 3D space.
    /// </summary>
    public class VDSprite : VDGameObject
    {
        private VDGraphics _graphics;           // graphics reference
        private Direct3D.Texture _texture;      // texture
        private string _textureName;            // texture's name
        private Direct3D.Sprite _sprite;        // the sprite!
        private bool _visible;                  // visibility
        private bool _transparent;              // transparency
        private SizeF _size;                    // texture size
        private Vector3 _position;              // location
        private Vector3 _center;                // rotation center
        private Vector3 _velocity;              // velocity
        private Color _color;                   // color
        private Rectangle _sourceRectangle;     // rectangle for the texture coordinates
        private Rectangle _collisionRectangle;  // rectangle for collision detection
        private Matrix _transform;              // sprite transform (scaling)

        /// <summary>
        /// Initializes the sprite.
        /// </summary>
        /// <param name="graphics">Owner of the sprite.</param>
        public VDSprite(VDGraphics graphics)
        {
            Direct3D.Surface surface;

            // TODO: Better texture loading, maybe from resources...
            // tex = TextureLoader.FromStream(Graphics.D3DDevice, GetType().Module.Assembly.GetManifestResourceStream(sTextureName));

            _graphics = graphics;                                                                           // set the graphics reference
            _texture = Direct3D.TextureLoader.FromFile(graphics.Device, "../../DavesDrawings.bmp");         // load the texture
            _textureName = string.Empty;                                                                    // no name
            surface = _texture.GetSurfaceLevel(0);                                                          // get texture dimensions
            _visible = true;                                                                                // sprite is visible by default
            _transparent = true;                                                                            // sprite is transparent by default
            _size = new SizeF(surface.Description.Width, surface.Description.Height);                       // original texture size
            _position = new Vector3(0.0f, 0.0f, 0.0f);                                                      // initial position to the upper-left corner of the screen
            _center = new Vector3(0.0f, 0.0f, 0.0f);                                                        // rotation center to the upper-left corner of the sprite
            _velocity = new Vector3(0.0f, 0.0f, 0.0f);                                                      // velocity is 0
            _color = Color.White;                                                                           // sprite is white
            _sourceRectangle = new Rectangle(0, 0, surface.Description.Width, surface.Description.Height);  // texture rectangle is size of texture
            _collisionRectangle = _sourceRectangle;                                                         // collision detection rectangle is same as texture rectangle
            _transform = Matrix.Identity;                                                                   // scaling is 1:1
        }

        /// <summary>
        /// Accessor for the visibility of the sprite.
        /// </summary>
        public bool Visible
        {
            get
            {
                return _visible;
            }
        }

        /// <summary>
        /// Accessor for the transparent choice of the sprite.
        /// </summary>
        public bool Transparent
        {
            get
            {
                return _transparent;
            }
        }

        /// <summary>
        /// Accessor for the collision detection rectangle.
        /// </summary>
        public Rectangle CollisionRectangle
        {
            get
            {
                // TODO: ?_?
                Rectangle newCollisionRectangle = new Rectangle((int) _position.X + _collisionRectangle.Left,
                                                                (int) _position.Y + _collisionRectangle.Top,
                                                                _collisionRectangle.Width,
                                                                _collisionRectangle.Height);
                return newCollisionRectangle;
            }
        }

        /// <summary>
        /// Scaling transformation calculations.
        /// </summary>
        public void ScalingTransformation()
        {
            float scaleX = 1.0f;
            float scaleY = 1.0f;

            if (_size.Width != _sourceRectangle.Width)
            {
                scaleX = (float) _size.Width / (float) _sourceRectangle.Width;
            }
            if (_size.Height != _sourceRectangle.Height)
            {
                scaleY = (float) _size.Height / (float) _sourceRectangle.Height;
            }
            _transform = Matrix.Scaling(scaleX, scaleY, 1.0f);  // scaleX * aspectRatio
        }

        /// <summary>
        /// Updates the sprite based on the elapsed time.
        /// </summary>
        /// <param name="elapsedMilliseconds">Milliseconds that have past since last cycle.</param>
        public void Update(double elapsedMilliseconds)
        {
        }

        /// <summary>
        /// Draws the sprite.
        /// </summary>
        public void Draw()
        {
            ScalingTransformation();
            _sprite.Transform = _transform;
            _sprite.Draw(_texture, _sourceRectangle, _center, _position, _color);
        }

        /// <summary>
        /// Releases the resources of the sprite..
        /// </summary>
        public void Shutdown()
        {
            if (_texture != null)
            {
                _texture.Dispose();
                _texture = null;
            }
        }
    }
}
