﻿using System;             // needed for Math class
using Microsoft.DirectX;  // needed for Vector3 structure
using VDEngine.DirectX;

namespace VDEngine.Game
{
    /// <summary>
    /// Player information.
    /// </summary>
    public class VDPlayer
    {
        private float _velocity;     // player's speed
        private Vector3 _direction;  // player's direction
        private Vector3 _location;   // player's position

        /// <summary>
        /// Direction of movement.
        /// </summary>
        public enum Movement
        {
            /// <summary>
            /// Moving forwards.
            /// </summary>
            Forward,
            /// <summary>
            /// Moving backwards.
            /// </summary>
            Backward,
            /// <summary>
            /// Strafing left.
            /// </summary>
            StrafeLeft,
            /// <summary>
            /// Strafing right.
            /// </summary>
            StrafeRight
        }

        /// <summary>
        /// Initializes the player.
        /// </summary>
        public VDPlayer()
        {
            _velocity = 1.0f;
            _direction = new Vector3(0.0f, 0.0f, 0.0f);
            _location = new Vector3(0.0f, 0.0f, 0.0f);
        }

        /// <summary>
        /// Accessor and modifier for the player's velocity.
        /// </summary>
        public float Velocity
        {
            get
            {
                return _velocity;
            }
            set
            {
                _velocity = value;
            }
        }

        /// <summary>
        /// Accessor for the player's direction.
        /// </summary>
        public Vector3 Direction
        {
            get
            {
                return _direction;
            }
        }

        /// <summary>
        /// Accessor for the player's location.
        /// </summary>
        public Vector3 Location
        {
            get
            {
                return _location;
            }
        }

        /// <summary>
        /// Updates the player's direction he or she is facing.
        /// </summary>
        /// <param name="mouseDirection">Mouse movement.</param>
        /// <param name="mouseSpeed">Speed of the mouse.</param>
        public void UpdateDirection(VDGraphics.MouseLook mouseDirection, int mouseSpeed)
        {
            // TODO: used to be based on time, but that sucks
            // _direction.X += (float) Math.Abs(mouseSpeed) * (float) elapsedTime / 10.0f;
            switch (mouseDirection)
            {
                case VDGraphics.MouseLook.Stasis:
                {
                    break;
                }
                case VDGraphics.MouseLook.Up:
                {
                    _direction.Y += (float) 0.01f * Math.Abs(mouseSpeed);
                    break;
                }
                case VDGraphics.MouseLook.Down:
                {
                    _direction.Y -= (float) 0.01f * Math.Abs(mouseSpeed);
                    break;
                }
                case VDGraphics.MouseLook.Left:
                {
                    _direction.X -= (float) 0.01f * Math.Abs(mouseSpeed);
                    break;
                }
                case VDGraphics.MouseLook.Right:
                {
                    _direction.X += (float) 0.01f * Math.Abs(mouseSpeed);
                    break;
                }
                default:
                {
                    break;
                }
            }
        }

        /// <summary>
        /// Move the player and update the location.
        /// </summary>
        /// <param name="movement">Direction of movement.</param>
        /// <param name="elapsedTime">Time past since the last game cycle.</param>
        public void Move(Movement movement, double elapsedTime)
        {
            switch (movement)
            {
                case Movement.Forward:
                {
                    _location.X += (float) Math.Sin(_direction.X) * _velocity * (float) elapsedTime / 100.0f;
                    _location.Z += (float) Math.Cos(_direction.X) * _velocity * (float) elapsedTime / 100.0f;
                    break;
                }
                case Movement.Backward:
                {
                    _location.X -= (float) Math.Sin(_direction.X) * _velocity * (float) elapsedTime / 100.0f;
                    _location.Z -= (float) Math.Cos(_direction.X) * _velocity * (float) elapsedTime / 100.0f;
                    break;
                }
                case Movement.StrafeLeft:
                {
                    _location.X -= (float) Math.Sin(_direction.X + Math.PI / 2) * _velocity * (float) elapsedTime / 100.0f;
                    _location.Z -= (float) Math.Cos(_direction.X + Math.PI / 2) * _velocity * (float) elapsedTime / 100.0f;
                    break;
                }
                case Movement.StrafeRight:
                {
                    _location.X += (float) Math.Sin(_direction.X + Math.PI / 2) * _velocity * (float) elapsedTime / 100.0f;
                    _location.Z += (float) Math.Cos(_direction.X + Math.PI / 2) * _velocity * (float) elapsedTime / 100.0f;
                    break;
                }
                default:
                {
                    break;
                }
            }
        }
    }
}
