﻿using System;                          // needed for Exception class
using System.Runtime.InteropServices;  // needed to import the DLL's

namespace VDEngine.Utility
{
    /// <summary>
    /// High-performance timer.
    /// </summary>
    public class VDTimer
    {
        /// <summary>
        /// Win32 API function that retrieves the number of counts per second in the high-resolution counter.
        /// </summary>
        /// <param name="lpFrequency"></param>
        /// <returns></returns>
        [DllImport("Kernel32.dll")]
        private static extern bool QueryPerformanceFrequency(ref long lpFrequency);

        /// <summary>
        /// Win32 API function that retrieves the number of high-resolution performance counters.
        /// </summary>
        /// <param name="lpPerformanceCount">Number of performance counters.</param>
        /// <returns></returns>
        [DllImport("Kernel32.dll")]
        private static extern bool QueryPerformanceCounter(ref long lpPerformanceCount);

        private long _lastTime;
        private long _currentTime;
        private long _ticksPerSecond;
        private double _elapsedMilliseconds;
        private int _frames;
        private int _framesLimit;
        private float _milliseconds;
        private float _millisecondsLimit;
        private float _framesPerSecond;
        private float _framesPerSecondLimit;
        private bool _fpsLimit;

        /// <summary>
        /// Initializes the the timer.
        /// </summary>
        public VDTimer()
        {
            _lastTime = 0;
            _currentTime = 0;
            _ticksPerSecond = 0;
            _elapsedMilliseconds = 0;
            _frames = 0;
            _framesLimit = 0;
            _milliseconds = 0.0f;
            _millisecondsLimit = 0.0f;
            _framesPerSecond = 0.0f;
            _framesPerSecondLimit = 0.0f;
            _fpsLimit = false;

            if (!QueryPerformanceFrequency(ref _ticksPerSecond))
            {
                throw new Exception("High-performance timer is not supported.");
            }
        }

        /// <summary>
        /// Accessor for the frames per second.
        /// </summary>
        public float FramesPerSecond
        {
            get
            {
                return _framesPerSecond;
            }
        }

        /// <summary>
        /// Starts the timer.
        /// </summary>
        public void Start()
        {
            QueryPerformanceCounter(ref _lastTime);
        }

        /// <summary>
        /// Calculates the number of milliseconds that have passed since the last call.
        /// </summary>
        public void CalculateElapsedMilliseconds()
        {
            // get the current number of ticks
            QueryPerformanceCounter(ref _currentTime);

            // get the number of milliseconds from the last method call
            _elapsedMilliseconds = ((double) (_currentTime - _lastTime) / (double) _ticksPerSecond) * 1000.0f;

            // set the current number of ticks for the next method call
            _lastTime = _currentTime;
        }

        /// <summary>
        /// Updates the game's FPS.
        /// </summary>
        /// <param name="targetFPS">The FPS that is desired.</param>
        public void UpdateFramesPerSecond(float targetFPS)
        {
            // calculate frames and milliseconds
            if (!_fpsLimit)
            {
                _milliseconds += (float) _elapsedMilliseconds;
                _frames++;

                _millisecondsLimit += (float) _elapsedMilliseconds;
                _framesLimit++;

                if (_milliseconds >= 1000.0f)
                {
                    // update frames per second
                    _framesPerSecond = (float) _frames / _milliseconds * 1000.0f;
                    _milliseconds = 0.0f;
                    _frames = 0;
                    //_fpsLimit = true;  // TODO: this fps limit thing... not working
                }
            }
            // hold it till we get target fps
            else 
            {
                _framesPerSecondLimit = (float) _framesLimit / _millisecondsLimit * 1000.0f;

                while (_framesPerSecondLimit > targetFPS)
                {
                    // TODO: this is wrong... redo the FPS calculation
                    CalculateElapsedMilliseconds();
                    _millisecondsLimit += (float) _elapsedMilliseconds;
                    _framesPerSecondLimit = (float) _framesLimit / _millisecondsLimit * 1000.0f;
                }

                _framesLimit = 0;
                _millisecondsLimit = 0;
                _fpsLimit = false;
            }
        }

        /// <summary>
        /// Accessor for the elapsed milliseconds since the last game loop.
        /// </summary>
        public double ElapsedMilliseconds
        {
            get
            {
                return _elapsedMilliseconds;
            }
        }
    }
}
