// BaseClass : DerivedClass     (I guess this is inheritance)
//
// KEYWORDS
// --------
// base         Refers to the base class
// this         Refers to the current instance of the class
//
// MEMBER ACCESS MODIFIERS
// -----------------------
// abstract     
// override     Extends or modifies the abstract or virtual implementation of a method, property, indexer, or event
// private      Accessible within its class or struct only
// protected    Accessible within its class and by derived class
// public       Accessible by all
// static       Does not need to be initialized (part of class, not object)
// virtual      Extends to subclasses
//
//
//               TODO LIST
//
// ~ Go through the entire program, make sure everything's tied to VDEngine nicely.
// ~ Work on the Particle System.
// ~ Figure out how to properly limit the game speed (Frames Per Second).
// ~ Options menu (this comes last)
//   May need to do a _device.Reset(_presentationParameters); when user changes graphic options
//   NOTE: If we do that, we want the player to still be in the same location and area as before!


using System;                          // needed for many things
using System.Drawing;                  // needed for Size structure
using System.Windows.Forms;            // needed for window things
using System.Runtime.InteropServices;  // needed to import the DLL
using Microsoft.DirectX.DirectInput;   // needed for keyboard stuff
using VDEngine.DirectX;
using VDEngine.Game;
using VDEngine.Utility;

namespace VDEngine
{
    /// <summary>
    /// The main game engine.
    /// </summary>
    public class VDEngine : System.Windows.Forms.Form
    {
        private VDDisplaySettings _userDisplaySettings;  // user's display options
        private VDPlayer _player;                        // player information
        private VDGraphics _graphics;                    // Direct3D graphics
        private VDTimer _timer;                          // application's high-performance timer
        private VDInput _input;                          // DirectInput input
        private VDAudio _audio;                          // DirectSound audio
        private VDParticleSystem _particleSystem;        // particle system
        private GameStates _gameState;                   // game state

        /// <summary>
        /// States of the game.
        /// </summary>
        public enum GameStates
        {
            /// <summary>
            /// Menu
            /// </summary>
            Menu,
            /// <summary>
            /// Game
            /// </summary>
            Game,
            /// <summary>
            /// Shutdown
            /// </summary>
            Exit
        }



        #region Initialization (Constructor)

        /// <summary>
        /// Initializes every aspect of the game.
        /// </summary>
        public VDEngine()
        {
            // initialize the user's display settings
            Console.WriteLine("Loading player's settings...");
            _userDisplaySettings = new VDDisplaySettings();

            // initialize the player
            Console.WriteLine("Loading player...");
            _player = new VDPlayer();

            // initialize the timer
            Console.WriteLine("Loading timer...");
            _timer = new VDTimer();

            // initialize the window
            Console.WriteLine("Loading the window...");
            InitializeWindow();

            // initialize the Direct3D graphics
            Console.WriteLine("Loading Direct3D...");
            _graphics = new VDGraphics(this);

            // initialize the DirectInput input devices
            Console.WriteLine("Loading input devices...");
            _input = new VDInput(this);

            // initialize DirectSound sound devices
            Console.WriteLine("Loading music and sounds...");
            _audio = new VDAudio(this);

            // initialize the particle system
            Console.WriteLine("Loading particle system...");
            _particleSystem = new VDParticleSystem();

            _gameState = GameStates.Game;
        }

        #endregion



        #region Window Stuff

        /// <summary>
        /// Initializes the window.
        /// </summary>
        private void InitializeWindow()
        {
            // setup window & full-screen properties
            AutoScaleMode = AutoScaleMode.None;                  // disable automatic scaling
            BackColor = System.Drawing.Color.Black;              // background color
            StartPosition = FormStartPosition.CenterScreen;      // center the initial position of the window
            Text = "Awww shit";                                  // window caption
            Cursor.Hide();                                       // hide the cursor
            SetStyle(ControlStyles.UserPaint, true);             // TODO: find out what this does (takes control of window rendering?)
            SetStyle(ControlStyles.AllPaintingInWmPaint, true);  // TODO: find out what this does (takes control of window rendering?)

            // start in windowed mode
            WindowedMode();
        }

        /// <summary>
        /// Set the window form to full-screen mode
        /// </summary>
        private void FullScreenMode()
        {
            // hide the window
            Hide();

            // make the necessary changes
            TopMost = true;                                   // set the window to be on top of all other windows
            ClientSize = (Screen.PrimaryScreen.Bounds).Size;  // set the window to encompass the whole screen (TODO: find out why it's base.ClientSize)
            FormBorderStyle = FormBorderStyle.None;           // remove the window's form style
            CenterToScreen();                                 // center the form on the screen (TODO: see if this is necessary)
            Cursor.Hide();                                    // hide the cursor

            // show the window
            Show();
        }

        /// <summary>
        /// Set the window to windowed mode
        /// </summary>
        private void WindowedMode()
        {
            // hide the window
            Hide();

            // make the necessary changes
            ClientSize = _userDisplaySettings.WindowSize;  // set the window to have the chosen dimensions
            FormBorderStyle = FormBorderStyle.Fixed3D;     // set the border style
            MinimizeBox = false;                           // disable the minimize icon
            MaximizeBox = false;                           // disable the maximize icon
            TopMost = false;                               // enable the window to be under other windows

            // show the window
            Show();
        }

        #endregion



        #region Windows Messaging

        /// <summary>
        /// Win32 API function to check for new messages.
        /// </summary>		
        [System.Security.SuppressUnmanagedCodeSecurity]
        [DllImport("User32.dll", CharSet = CharSet.Auto)]
        private static extern bool PeekMessage(out ThePeekMessage msg, IntPtr hWnd, uint messageFilterMin, uint messageFilterMax, uint flags);

        /// <summary>
        /// Windows message structure
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct ThePeekMessage
        {
            /// <summary>
            /// This variable is not used.
            /// </summary>
            public IntPtr handleWindow;
            /// <summary>
            /// This variable is not used.
            /// </summary>
            public Message message;
            /// <summary>
            /// This variable is not used.
            /// </summary>
            public IntPtr wParameter;
            /// <summary>
            /// This variable is not used.
            /// </summary>
            public IntPtr lParameter;
            /// <summary>
            /// This variable is not used.
            /// </summary>
            public uint time;
            /// <summary>
            /// This variable is not used.
            /// </summary>
            public System.Drawing.Point point;
        }

        /// <summary>
        /// Checks for events in the queue and runs the main game loop if there are none.
        /// </summary>
        /// <param name="sender">Sender of the event.  This parameter is not used.</param>
        /// <param name="eventArguments">Arguments of the event.  This parameter is not used.</param>
        private void OnApplicationIdle(object sender, EventArgs eventArguments)
        {
            ThePeekMessage message;

            // don't do anything if the user's not in this window
            if (ActiveForm != this || !this.Focused)
            {
                // TODO: Maybe Thread.Sleep(x)?
                return;
            }

            // see if there are no messages in the queue
            while (!PeekMessage(out message, IntPtr.Zero, 0, 0, 0))
            {
                GameLoop();
            }
        }

        #endregion



        #region Game Loop

        /// <summary>
        /// The main game loop.
        /// </summary>
        private void GameLoop()
        {
            // get elapsed time
            _timer.CalculateElapsedMilliseconds();
            _timer.UpdateFramesPerSecond(30.0f);

            // handle lost & reset graphics device
            if (_graphics.DeviceLost)
            {
                _audio.Shutdown();
                _input.Shutdown();
                _graphics.Shutdown();
            }
            if (_graphics.DeviceReset)
            {
                _graphics = new VDGraphics(this);
                _input = new VDInput(this);
                _audio = new VDAudio(this);
            }

            // get player input
            _input.UpdateKeyboard();
            _input.UpdateMouse();

            // perform ai and game logic
            switch (_gameState)
            {
                case GameStates.Game:
                {
                    // handle player input
                    HandleInput(_timer.ElapsedMilliseconds);
                    break;
                }
                case GameStates.Exit:
                {
                    // shutdown gracefully
                    Shutdown();
                    return;
                }
                default:
                {
                    break;
                }
            }

            // handle sound (TODO: is this in the right spot?)
            _audio.SoundEffectManagement();

            // render
            _graphics.CalculateNewWorldView();  // new world perspective if user moves
            _graphics.Render();                 // graphics

            // synchronize display
            //while (!_timer.WaitUntilTargetFPS(30.0f))
            //{
                // Hahaha
            //}
        }

        /// <summary>
        /// Keyboard and mouse input based on last game loop.
        /// </summary>
        /// <param name="elapsedSeconds">Elapsed seconds from the last game loop.</param>
        private void HandleInput(double elapsedSeconds)
        {
            // user moves forward
            if (_input.KeyHeldDown(Key.W))
            {
                _player.Move(VDPlayer.Movement.Forward, elapsedSeconds);
            }

            // user moves backward
            else if (_input.KeyHeldDown(Key.S))
            {
                _player.Move(VDPlayer.Movement.Backward, elapsedSeconds);
            }

            // user strafes left
            else if (_input.KeyHeldDown(Key.A))
            {
                _player.Move(VDPlayer.Movement.StrafeLeft, elapsedSeconds);
            }

            // user strafes right
            else if (_input.KeyHeldDown(Key.D))
            {
                _player.Move(VDPlayer.Movement.StrafeRight, elapsedSeconds);
            }

            // see if the mouse moved
            switch (_input.MouseMoved)
            {
                // user did nothing
                case VDGraphics.MouseLook.Stasis:
                {
                    break;
                }
                // user looks up
                case VDGraphics.MouseLook.Up:
                {
                    _player.UpdateDirection(VDGraphics.MouseLook.Up, _input.MouseSpeed);
                    break;
                }
                // user looks down
                case VDGraphics.MouseLook.Down:
                {
                    _player.UpdateDirection(VDGraphics.MouseLook.Down, _input.MouseSpeed);
                    break;
                }
                // user looks left
                case VDGraphics.MouseLook.Left:
                {
                    _player.UpdateDirection(VDGraphics.MouseLook.Left, _input.MouseSpeed);
                    break;
                }
                // user looks right
                case VDGraphics.MouseLook.Right:
                {
                    _player.UpdateDirection(VDGraphics.MouseLook.Right, _input.MouseSpeed);
                    break;
                }
                default:
                {
                    break;
                }
            }

            // TODO: temporary bullet test
            if (_input.KeyHeldDown(Key.F))
            {
                _particleSystem.FireBullets(_player);
            }

            // TODO: temporary sound test
            if (_input.KeyHeldDown(Key.P))
            {
                _audio.TempOnClick();
            }

            // end game
            if (_input.KeyHeldDown(Key.Escape))
            {
                _gameState = GameStates.Exit;
            }
        }

        #endregion



        #region Main
        /// <summary>
        /// The main entry point for the engine.  The COM threading model for this application is single-threaded apartment.
        /// </summary>
        [STAThread]
        static void Main()
        {
            // initialize the engine
            VDEngine game = new VDEngine();
            Console.WriteLine("\nLoading successful.  Proceeding to Game Loop.");

            // add the game's idle event
            System.Windows.Forms.Application.Idle += new EventHandler(game.OnApplicationIdle);

            // begin running a standard application message loop on this thread, and make the game window visible
            System.Windows.Forms.Application.Run(game);
        }
        #endregion



        #region Shutdown

        /// <summary>
        /// Releases the resources of the engine.
        /// </summary>
        public void Shutdown()
        {
            _audio.Shutdown();
            _input.Shutdown();
            _graphics.Shutdown();
            Application.Exit();
        }

        #endregion



        #region Accessors & Modifiers

        /// <summary>
        /// Accessor for the user display settings.
        /// </summary>
        public VDDisplaySettings DisplaySettings
        {
            get
            {
                return _userDisplaySettings;
            }
        }

        /// <summary>
        /// Accessor for player information.
        /// </summary>
        public VDPlayer Player
        {
            get
            {
                return _player;
            }
        }

        /// <summary>
        /// Accessor for the game's timer.
        /// </summary>
        public VDTimer Timer
        {
            get
            {
                return _timer;
            }
        }

        /// <summary>
        /// Accessor for the particle system.
        /// </summary>
        public VDParticleSystem ParticleSystem
        {
            get
            {
                return _particleSystem;
            }
        }

        #endregion
    }
}
